__all__ = [
    "ForeignKey",
    "Relationship",
    "Query", "Session",
    "SubtypesDateTime", "SubtypesDate", "BitLiteral",
    "ModelMeta", "Model", "AutoModel", "ReflectedModel",
    "Table",
    "Select", "Update", "Insert", "Delete", "SelectInto",
]

from .misc import ForeignKey
from .relationship import Relationship
from .query import Query, Session
from .field import SubtypesDateTime, SubtypesDate, BitLiteral
from .model import ModelMeta, Model, AutoModel, ReflectedModel
from .table import Table
from .expression import Select, Update, Insert, Delete, SelectInto



from __future__ import annotations

from typing import Any

from django import db

from subtypes import Dict_

from .config import SqlConfig
from .database import DjangoDatabase, DjangoApps
from .model import SqlModel

from sqlhandler.utils import Url


class DjangoSql(SqlConfig.Sql):
    class Settings(SqlConfig.Sql.Settings):
        eager_reflection = cache_metadata = True
        reflect_views = False

    class Constructors(SqlConfig.Sql.Constructors):
        ReflectedModel, Database = SqlModel, DjangoDatabase

    @property
    def django(self) -> DjangoApps:
        return self.database.django

    def _create_url(self, connection: str, **kwargs: Any) -> Url:
        detail = Dict_(db.connections.databases[connection])
        drivername = SqlConfig.settings.ENGINES[detail.ENGINE.rpartition(".")[-1]]
        return Url(drivername=drivername, database=detail.NAME, username=detail.USER or None, password=detail.PASSWORD or None, host=detail.HOST or None, port=detail.PORT or None)



from __future__ import annotations

from typing import Type, TYPE_CHECKING

from django.db.models.manager import Manager
from django.db.models.options import Options

from .config import SqlConfig

if TYPE_CHECKING:
    from .sql import DjangoSql


class SqlModel(SqlConfig.Sql.Constructors.ReflectedModel, SqlConfig.settings.MODEL_MIXIN):
    @classmethod
    def django(cls) -> Type[DjangoModel]:
        return SqlConfig.sql.database.django_mappings[cls.__table__.name]

    @classmethod
    def handler(cls) -> DjangoSql:
        return SqlConfig.sql

    def __call__(self) -> DjangoModel:
        return type(self).django().objects.get(pk=getattr(self, list(self.__table__.primary_key)[0].name))


class DjangoModel:
    objects: Manager
    _meta: Options

    @classmethod
    def sql(cls) -> Type[SqlModel]:
        return SqlConfig.sql.database.sqlhandler_mappings[cls._meta.db_table]

    def __call__(self) -> SqlModel:
        return type(self).sql().query.get(getattr(self, self._meta.pk.name))



__all__ = ["SqlConfig", "DjangoModel", "DjangoSql", "sql"]

from typing import Optional

from .config import SqlConfig
from .model import DjangoModel
from .sql import DjangoSql

sql: Optional[DjangoSql] = None



from __future__ import annotations

from typing import Optional


class ObjectName:
    object_type = "object"

    def __init__(self, stem: str, schema: SchemaName) -> None:
        self.stem, self.schema, self.name = stem, schema, stem if schema.name is None else f"{schema.name}.{stem}"

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.stem

    def __eq__(self, other: ObjectName) -> bool:
        return self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class TableName(ObjectName):
    object_type = "table"


class ViewName(ObjectName):
    object_type = "view"


class SchemaName:
    def __init__(self, name: Optional[str], default: str) -> None:
        self.name, self.default = name if name is not None else (None if default is None else default), default
        self.is_default = name is None or name == default

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: [SchemaName, str]) -> bool:
        if isinstance(other, type(self)):
            return self.name == other.name
        else:
            return True if other is None and self.is_default else self.name == other

    def __hash__(self) -> int:
        return hash(self.name)




from __future__ import annotations

from typing import TYPE_CHECKING, Set, Dict, Union

from miscutils import ReprMixin

from .name import SchemaName, ObjectName, TableName, ViewName

if TYPE_CHECKING:
    from .database import Database


class DatabaseShape(ReprMixin):
    def __init__(self, database: Database) -> None:
        self.database = database
        self.schemas: Dict[SchemaName, SchemaShape] = {name: SchemaShape(name=name, database=self.database) for name in self.database.schema_names()}
        self.schema_name_mappings = {name.name: name for name in self.schemas}

        self.all_objects = set()

    def __getitem__(self, name: Union[str, SchemaName]) -> SchemaShape:
        return self.schemas[name]

    def refresh(self):
        for schema in self.schemas.values():
            schema.refresh()

        self.all_objects: Set[ObjectName] = {obj for schema in self.schemas.values() for obj in schema.objects}


class SchemaShape(ReprMixin):
    def __init__(self, name: SchemaName, database: Database) -> None:
        self.name, self.database, self.registry = name, database, {}

    def refresh(self) -> None:
        self.tables: Set[TableName] = self.database.table_names(self.name)
        self.views: Set[ViewName] = self.database.view_names(self.name)
        self.objects: Set[ObjectName] = self.tables | self.views



from __future__ import annotations

from typing import Any

import sqlalchemy as alch
from sqlalchemy.sql.schema import _get_table_key


class Table(alch.Table):
    def __new__(*args: Any, **kwargs: Any) -> Table:
        if len(args) == 1:
            return alch.Table.__new__(*args, **kwargs)

        _, name, meta, *_ = args
        if kwargs.pop("is_declarative", False):
            if (schema := kwargs.get("schema")) is None:
                schema = meta.schema

            if (table := meta.tables.get(_get_table_key(name, schema))) is not None:
                meta.remove(table)

        return alch.Table.__new__(*args, **kwargs)



from __future__ import annotations

import warnings
from contextlib import contextmanager
from typing import Any, Union, Set, Callable, TYPE_CHECKING, cast, Type

import sqlalchemy as alch
from sqlalchemy import Column, Integer, event
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base

from maybe import Maybe
from subtypes import Str, Dict_
from miscutils import cached_property, PercentagePrinter, Printer
from iotools import Cache

from .meta import NullRegistry, Metadata
from .name import SchemaName, ObjectName, ViewName, TableName
from .shape import DatabaseShape
from .schema import Schemas, TableSchemas, ViewSchemas, SchemaRouter

from sqlhandler.custom import Model, AutoModel, Table

if TYPE_CHECKING:
    from sqlhandler import Sql


class Database:
    """A class representing a sql database. Abstracts away database reflection and metadata caching. The cache lasts for 5 days but can be cleared with Database.clear()"""
    _null_registry = NullRegistry()

    def __init__(self, sql: Sql) -> None:
        self.sql, self.name, self.cache, self._post_reshape_countdown = sql, sql.engine.url.database, Cache(file=sql.config.folder.new_file("sql_cache", "pkl"), days=5), 0
        self.meta = self._get_metadata()

        self.model = cast(Type[Model], declarative_base(metadata=self.meta, cls=self.sql.constructors.Model, metaclass=self.sql.constructors.ModelMeta, name=self.sql.constructors.Model.__name__, class_registry=self._null_registry))
        self.auto_model = cast(Type[AutoModel], declarative_base(metadata=self.meta, cls=self.sql.constructors.AutoModel, metaclass=self.sql.constructors.ModelMeta, name=self.sql.constructors.AutoModel.__name__, class_registry=self._null_registry))

        self.shape = DatabaseShape(database=self)
        self.objects, self.tables, self.views = Schemas(database=self), TableSchemas(database=self), ViewSchemas(database=self)
        self.router = SchemaRouter(database=self).register_all([self.objects, self.tables, self.views])

        self._sync_with_db()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.name)}, schemas={len(self.shape.schemas)}, tables={len(self.meta.tables)})"

    def __call__(self) -> Database:
        self._reflect_database()
        return self

    @cached_property
    def default_schema(self) -> str:
        if name := alch.inspect(self.sql.engine).default_schema_name:
            return name
        else:
            name, = alch.inspect(self.sql.engine).get_schema_names() or [None]
            return name

    def schema_names(self) -> Set[SchemaName]:
        return {SchemaName(name=name, default=self.default_schema) for name in alch.inspect(self.sql.engine).get_schema_names()}

    def table_names(self, schema: SchemaName) -> Set[TableName]:
        return {TableName(stem=name, schema=schema) for name in alch.inspect(self.sql.engine).get_table_names(schema=schema.name)}

    def view_names(self, schema: SchemaName) -> Set[ViewName]:
        return {ViewName(stem=name, schema=schema) for name in alch.inspect(self.sql.engine).get_view_names(schema=schema.name)}

    def create_table(self, table: Table) -> None:
        """Emit a create table statement to the database from the given table object."""
        table = self._normalize_table(table)
        table.create()
        self._sync_with_db()
        self._reflect_object(self._name_from_object(table, object_type=TableName))

    def drop_table(self, table: Table) -> None:
        """Emit a drop table statement to the database for the given table object."""
        table = self._normalize_table(table)
        table.drop()
        self._sync_with_db()

    def refresh_table(self, table: Table) -> None:
        """Reflect the given table object again."""
        table = self._normalize_table(table)
        self._sync_with_db()
        self._reflect_object(self._name_from_object(table, object_type=TableName))

    def exists_table(self, table: Table) -> bool:
        table = self._normalize_table(table)
        with self.sql.engine.connect() as con:
            return self.sql.engine.dialect.has_table(con, table.name, schema=table.schema)

    def reset(self) -> None:
        """Clear this database's metadata as well as its cache and reflect everything from scratch."""
        self.meta.clear()
        self._cache_metadata()
        self._sync_with_db()

    def _get_metadata(self) -> Metadata:
        if not self.sql.settings.cache_metadata:
            return self.sql.constructors.Metadata(sql=self.sql, bind=self.sql.engine)

        try:
            meta = self.cache.setdefault(self.name, self.sql.constructors.Metadata())
        except Exception as ex:
            warnings.warn(f"The following exception ocurred when attempting to retrieve the previously cached Metadata, but was supressed:\n\n{ex}\n\nStarting with blank Metadata...")
            meta = self.sql.constructors.Metadata()

        meta.bind, meta.sql = self.sql.engine, self.sql

        return meta

    def _cache_metadata(self) -> None:
        if self.sql.settings.cache_metadata:
            self.cache[self.name] = self.meta

    @contextmanager
    def _post_reshape_soon(self) -> None:
        self._post_reshape_countdown += 1
        yield
        self._post_reshape_countdown -= 1

        if not self._post_reshape_countdown:
            self._cache_metadata()
            self._autoload_models()

    def _sync_with_db(self) -> None:
        with self._post_reshape_soon():
            self.shape.refresh()
            self.router.refresh_accessors()

            if not self.meta.tables and self.sql.settings.eager_reflection:
                self._reflect_database()

            self._remove_expired_metadata_objects()

    def _reflect_database(self):
        with self._post_reshape_soon():
            for schema in PercentagePrinter(sorted(self.shape.schemas, key=lambda name: name.name), formatter=lambda name: f"Reflecting schema: {name.name}"):
                if schema.name not in self.sql.settings.lazy_schemas:
                    with Printer.from_indentation():
                        self._reflect_schema(schema=schema)

    def _reflect_schema(self, schema: SchemaName):
        with self._post_reshape_soon():
            schema_shape = self.shape[schema]
            names = sum([
                sorted(collection, key=lambda name_: name_.name) if condition else []
                for collection, condition in [(schema_shape.tables, self.sql.settings.reflect_tables), (schema_shape.views, self.sql.settings.reflect_views)]
            ], [])

            for name in PercentagePrinter(names, formatter=lambda name_: f"Reflecting {name_.object_type}: {name_.name}"):
                self._reflect_object(name=name)

    # noinspection PyArgumentList
    def _reflect_object(self, name: ObjectName) -> Table:
        with self._post_reshape_soon():
            if not (table := Table(name.stem, self.meta, schema=name.schema.name, autoload=True)).primary_key:
                table = Table(name.stem, self.meta, Column("__pk__", Integer, primary_key=True), schema=name.schema.name, autoload=True, extend_existing=True)

            return table

    def _autoload_models(self) -> None:
        reflected_model = declarative_base(bind=self.sql.engine, metadata=self.meta, cls=self.sql.constructors.ReflectedModel, metaclass=self.sql.constructors.ModelMeta, name=self.sql.constructors.ReflectedModel.__name__, class_registry=(registry := {}))
        automap = automap_base(declarative_base=reflected_model)
        automap.classes = Dict_()

        @event.listens_for(automap, "class_instrument", propagate=True)
        def cls_instrument(cls):
            schema = schema if (schema := (table := cls.__table__).schema) is not None else cls.metadata.sql.database.default_schema
            automap.classes.setdefault_lazy(key=schema, factory=Dict_)[table.name] = cls

        automap.prepare(classname_for_table=self._table_name(), name_for_scalar_relationship=self._scalar_name(), name_for_collection_relationship=self._collection_name())

        for key, val in automap.classes.items():
            if isinstance(val, Dict_):
                self.shape.schemas[key].registry.update(val)

    def _remove_expired_metadata_objects(self):
        with self._post_reshape_soon():
            for item in list(self.meta.tables.values()):
                if self._name_from_object(item) not in self.shape.all_objects:
                    self._remove_object_if_exists(item)

    def _remove_object_if_exists(self, table: Table) -> None:
        if table in self.meta:
            with self._post_reshape_soon():
                self.meta.remove(table)

        self.router.remove_object_if_exists(name=self._name_from_object(table=table, object_type=TableName))

    def _name_from_object(self, table: Table, object_type: Type[ObjectName] = ObjectName) -> ObjectName:
        return object_type(stem=table.name, schema=SchemaName(table.schema, default=self.default_schema))

    def _normalize_table(self, table: Union[Model, Table, str]) -> Table:
        return self.meta.tables[table] if isinstance(table, str) else Maybe(table).__table__.else_(table)

    def _table_name(self) -> Callable:
        def table_name(base: Any, tablename: Any, table: Any) -> str:
            return tablename

        return table_name

    def _scalar_name(self) -> Callable:
        def scalar_name(base: Any, local_cls: Any, referred_cls: Any, constraint: Any) -> str:
            return referred_cls.__name__

        return scalar_name

    def _collection_name(self) -> Callable:
        def collection_name(base: Any, local_cls: Any, referred_cls: Any, constraint: Any) -> str:
            return str(Str(referred_cls.__name__).case.snake().case.plural())

        return collection_name



from __future__ import annotations

from typing import Any, TYPE_CHECKING, Optional

import sqlalchemy as alch
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.ext.compiler import compiles

from maybe import Maybe
from subtypes import Frame
from miscutils import is_non_string_iterable

from sqlhandler.utils.utils import literal_statement


if TYPE_CHECKING:
    from sqlhandler import Sql


class ExpressionMixin:
    """A mixin providing private methods for logging using expression classes."""
    bind: Any
    table: Any
    select: Any
    into: Any
    _whereclause: Any

    @property
    def sql(self) -> Sql:
        return self.bind.sql

    def execute(self, autocommit: bool = False) -> str:
        """Execute this query's statement in the current session."""
        res = self.sql.session.execute(self)
        if autocommit:
            self.sql.session.commit()
        return res

    def _prepare_tran(self) -> None:
        self.sql.session.rollback()
        self.sql.log.write_sql(f"{'-' * 200}\n\nBEGIN TRAN;", add_newlines=2)

    def _resolve_tran(self, force_commit: bool = False) -> None:
        if self.sql.autocommit or force_commit:
            self.sql.session.commit()
            self.sql.log.write_sql("COMMIT;", add_newlines=2)
        else:
            user_confirmation = input("\nIf you are happy with the above Query/Queries please type COMMIT. Anything else will roll back the ongoing Transaction.\n\n")
            if user_confirmation.upper() == "COMMIT":
                self.sql.session.commit()
                self.sql.log.write_sql("COMMIT;", add_newlines=2)
            else:
                self.sql.session.rollback()
                self.sql.log.write_sql("ROLLBACK;", add_newlines=2)

    def _perform_pre_select(self, silently: bool) -> Optional[Select]:
        if silently:
            return
        else:
            pre_select_object = self.sql.Select(["*"]).from_(self.table)

            if self._whereclause is not None:
                pre_select_object = pre_select_object.where(self._whereclause)

            (pre_select_object.frame if silently else pre_select_object.resolve)()
            return pre_select_object

    def _perform_post_select(self, pre_select_object: Select, silently: bool) -> None:
        if not silently:
            pre_select_object.resolve()

    def _perform_pre_select_from_select(self, silently: bool) -> int:
        return None if self.select is None else len((self.select.frame if silently else self.select.resolve)().index)

    def _execute_expression_and_determine_rowcount(self, rowcount: int = None) -> int:
        result = self.sql.session.execute(self)
        self.sql.log.write_sql(str(self), add_newlines=2)

        if rowcount is None:
            rowcount = result.rowcount

        if rowcount == -1:
            if isinstance(self, Insert):
                if self.select is None:
                    rowcount = len(self.parameters) if isinstance(self.parameters, list) else 1

        self.sql.log.write_comment(f"({rowcount} row(s) affected)", add_newlines=2)

        return rowcount

    def _perform_post_select_inserts(self, rowcount: int, silently: bool) -> None:
        if not silently:
            self.sql.Select(["*"]).select_from(self.table).order_by(getattr(self.table.columns, list(self.table.primary_key)[0].name).desc()).limit(rowcount).resolve()

    def _perform_post_select_all(self, silently: bool) -> None:
        if not silently:
            self.sql.Select(["*"]).select_from(self.sql.text(f"{self.into}")).resolve()


class Select(alch.sql.Select):
    """Custom subclass of sqlalchemy.sql.Select with additional useful methods and aliases for existing methods."""

    def __init__(self, *columns: Any, whereclause: Any = None, from_obj: Any = None, distinct: Any = False, having: Any = None, correlate: Any = True, prefixes: Any = None, suffixes: Any = None, **kwargs: Any) -> None:
        as_single_iterable = columns[0] if len(columns) == 1 and is_non_string_iterable(columns[0]) else [*columns]
        super().__init__(columns=as_single_iterable, whereclause=whereclause, from_obj=from_obj, distinct=distinct, having=having, correlate=correlate, prefixes=prefixes, suffixes=suffixes, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def frame(self) -> Frame:
        """Execute the query and return the result as a subtypes.Frame."""
        return self._select_to_frame()

    def resolve(self) -> Frame:
        """Convert this query into a subtypes.Frame and write an ascii representation of it to the log, then return it."""
        frame = self._select_to_frame()
        self.bind.sql.log.write_sql(str(self), add_newlines=2)
        self.bind.sql.log.write_comment(frame.applymap(lambda val: 1 if val is True else (0 if val is False else ("NULL" if val is None else val))).to_ascii(), add_newlines=2)
        return frame

    def from_(self, *args: Any, **kwargs: Any) -> Select:
        """Simple alias for the 'select_from' method. See that method's docstring for documentation."""
        return self.select_from(*args, **kwargs)

    def _select_to_frame(self) -> Frame:
        result = self.bind.sql.session.execute(self)
        cols = [col[0] for col in result.cursor.description]
        return Frame(result.fetchall(), columns=cols)


class Update(ExpressionMixin, alch.sql.Update):
    """Custom subclass of sqlalchemy.sql.Update with additional useful methods and aliases for existing methods."""

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def resolve(self, silently: bool = False) -> None:
        """Execute this statement with surrounding Select statements as applicable, and request user confirmation to commit if Sql.autocommit is False, else commit the transaction."""
        self._prepare_tran()
        pre_select_object = self._perform_pre_select(silently=silently)
        self._execute_expression_and_determine_rowcount()
        self._perform_post_select(pre_select_object=pre_select_object, silently=silently)
        self._resolve_tran()

    def set_(self, *args: Any, **kwargs: Any) -> Update:
        """Simple alias for the 'values' method. See that method's docstring for documentation."""
        return self.values(*args, **kwargs)


class Insert(ExpressionMixin, alch.sql.Insert):
    """Custom subclass of sqlalchemy.sql.Insert with additional useful methods and aliases for existing methods."""

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def resolve(self, silently: bool = False) -> None:
        """Execute this statement with surrounding Select statements as applicable, and request user confirmation to commit if Sql.autocommit is False, else commit the transaction."""
        self._prepare_tran()
        rowcount = self._perform_pre_select_from_select(silently=silently)
        rowcount = self._execute_expression_and_determine_rowcount(rowcount=rowcount)
        self._perform_post_select_inserts(rowcount=rowcount, silently=silently)
        self._resolve_tran()

    def values(self, *args: Any, **kwargs: Any) -> Insert:
        """Insert the given values as either a single dict, or a list of dicts."""
        ret = super().values(*args, **kwargs)
        if isinstance(ret.parameters, list):
            ret.parameters = [{(col.key if isinstance(col, InstrumentedAttribute) else col): (Maybe(val).else_(alch.null()))
                               for col, val in record.items()} for record in ret.parameters]
        return ret


class Delete(ExpressionMixin, alch.sql.Delete):
    """Custom subclass of sqlalchemy.sql.Delete with additional useful methods and aliases for existing methods."""

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def resolve(self, silently: bool = False) -> None:
        """Execute this statement with surrounding Select statements as applicable, and request user confirmation to commit if Sql.autocommit is False, else commit the transaction."""
        self._prepare_tran()
        self._perform_pre_select(silently=silently)
        self._execute_expression_and_determine_rowcount()
        self._resolve_tran()


class SelectInto(ExpressionMixin, alch.sql.Select):
    """Custom subclass of sqlalchemy.sql.Select for 'SELECT * INTO #tmp' syntax with additional useful methods and aliases for existing methods."""

    def __init__(self, columns: list, *args: Any, table: str = None, schema: str = None, **kwargs: Any) -> None:
        self.into = f"{schema or 'dbo'}.{table}"
        super().__init__(columns, *args, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(\n{(str(self))}\n)"

    def __str__(self) -> str:
        return literal_statement(self)

    def resolve(self, silently: bool = False) -> None:
        """Execute this statement with surrounding Select statements as applicable, and request user confirmation to commit if Sql.autocommit is False, else commit the transaction."""
        self._prepare_tran()
        self._execute_expression_and_determine_rowcount()
        self._perform_post_select_all(silently=silently)
        self._resolve_tran(force_commit=True)


@compiles(SelectInto)  # type:ignore
def s_into(element: Any, compiler: Any, **kw: Any) -> Any:
    text = compiler.visit_select(element, **kw)
    text = text.replace("FROM", f"INTO {element.into} \nFROM")
    return text



from __future__ import annotations

from typing import Any, List, Callable, TypeVar, TYPE_CHECKING, Dict, Tuple, Optional
from abc import ABC, abstractmethod

from sqlalchemy.orm import Query
import sqlparse

from subtypes import Frame, Str
from pathmagic import File, PathLike

if TYPE_CHECKING:
    from sqlhandler import Sql


SelfType = TypeVar("SelfType")

# TODO: fix ON clause whitespace in all situations


class SqlBoundMixin:
    """A mixin class for objects that require a reference to an Sql object in their constructor."""

    def __init__(self, *args: Any, sql: Sql = None, **kwargs: Any) -> None:
        self.sql = sql

    @classmethod
    def from_sql(cls: SelfType, sql: Sql) -> Callable[[...], SelfType]:
        def wrapper(*args: Any, **kwargs: Any) -> SqlBoundMixin:
            return cls(*args, sql=sql, **kwargs)
        return wrapper


class Executable(SqlBoundMixin, ABC):
    """An abstract class representing a SQL executable such. Concrete implementations such as scripts or stored procedures must inherit from this. An implementaion of Executable._compile_sql() must be provided."""

    def __init__(self, sql: Sql = None, verbose: bool = False) -> None:
        self.sql = sql
        self.results: List[List[Frame]] = []

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.execute(*args, **kwargs)

    def execute(self, *args: Any, **kwargs: Any) -> Optional[List[Frame]]:
        """Execute this executable SQL object. Passes on its args and kwargs to Executable._compile_sql()."""
        statement, bindparams = self._compile_sql(*args, **kwargs)
        bound = statement

        if bindparams:
            for key, val in bindparams.items():
                bound = bound.replace(f":{key}", repr(val))

        self.sql.log.write_sql(bound)

        if (cursor := self.sql.session.execute(statement, bindparams).cursor) is None:
            return None
        else:
            self.results.append(self._get_frames_from_cursor(cursor))
            return self.results[-1]

    @abstractmethod
    def _compile_sql(self, *args: Any, **kwargs: Any) -> Tuple[str, Dict[str, Any]]:
        pass

    @staticmethod
    def _get_frames_from_cursor(cursor: Any) -> List[Frame]:
        def get_frame_from_cursor(curs: Any) -> Optional[Frame]:
            try:
                return Frame([tuple(row) for row in curs.fetchall()], columns=[info[0] for info in cursor.description])
            except Exception:
                return None

        data = [get_frame_from_cursor(cursor)]
        while cursor.nextset():
            data.append(get_frame_from_cursor(cursor))

        return [frame for frame in data if frame is not None] or None


class StoredProcedure(Executable):
    """A class representing a stored procedure in the database. Can be called to execute the proc. Arguments and keyword arguements will be passed on."""

    def __init__(self, name: str, schema: str = None, database: str = None, sql: Sql = None) -> None:
        super().__init__(sql=sql)
        self.name, self.schema, self.database = name, schema or self.sql.database.default_schema, database

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={self.name}, schema={self.schema})"

    def _compile_sql(self, *args: Any, **kwargs: Any) -> Tuple[str, Dict[str, Any]]:
        mappings = {
            **{f"arg{index + 1}": {"bind": f":arg{index + 1}", "val": val} for index, val in enumerate(args)},
            **{f"kwarg{index + 1}": {"bind": f"@{name}=:kwarg{index + 1}", "val": val} for index, (name, val) in enumerate(kwargs.items())}
        }
        proc_name = f"EXEC {f'[{self.database}].' if self.database is not None else ''}[{self.schema}].[{self.name}]"
        return f"{proc_name} {', '.join([arg['bind'] for arg in mappings.values()])}", {name: arg["val"] for name, arg in mappings.items()}


class Script(Executable):
    """A class representing a SQL script in the filesystem. Can be called to execute the script."""

    def __init__(self, path: PathLike, sql: Sql = None) -> None:
        super().__init__(sql=sql)
        self.file = File.from_pathlike(path)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(file={self.file})"

    def _compile_sql(self, *args: Any, **kwargs: Any) -> Tuple[str, Dict[str, Any]]:
        return self.file.content, {}


def literal_statement(statement: Any, format_statement: bool = True) -> str:
    """Returns this a query or expression object's statement as raw SQL with inline literal binds."""

    if isinstance(statement, Query):
        statement = statement.statement

    bound = statement.compile(compile_kwargs={'literal_binds': True}).string + ";"
    formatted = sqlparse.format(bound, reindent=True) if format_statement else bound  # keyword_case="upper" (removed arg due to false positives)

    stage1 = Str(formatted).re.sub(r"\bOVER\s*\(\s*", lambda m: "OVER (").re.sub(r"OVER \((ORDER\s*BY|PARTITION\s*BY)\s+(\S+)\s+(ORDER\s*BY|PARTITION\s*BY)\s+(\S+)\s*\)", lambda m: f"OVER ({m.group(1)} {m.group(2)} {m.group(3)} {m.group(4)})")
    stage2 = stage1.re.sub(r"(?<=\n)([^\n]*JOIN[^\n]*)(\bON\b[^\n;]*)(?=[\n;])", lambda m: f"  {m.group(1).strip()}\n    {m.group(2).strip()}")
    stage3 = stage2.re.sub(r"(?<=\bJOIN[^\n]+\n\s+ON[^\n]+\n(?:\s*AND[^\n]+\n)*?)(\s*AND[^\n]+)(?=[\n;])", lambda m: f"    {m.group(1).strip()}")

    return str(stage3)



from __future__ import annotations

from typing import Any, TYPE_CHECKING, List, Collection

from subtypes import NameSpace

from .name import SchemaName, ObjectName

from sqlhandler.custom import Model

if TYPE_CHECKING:
    from .database import Database
    from .shape import DatabaseShape, SchemaShape


class BaseSchema(NameSpace):
    """A NameSpace class representing a database schema. Models/objects can be accessed with either attribute or item access. If the model/object isn't already cached, an attempt will be made to reflect it."""

    def __init__(self, name: SchemaName, parent: Schemas) -> None:
        self._name, self._parent, self._database = name, parent, parent._database

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name='{self._name}', num_objects={len(self)}, objects={[table for table, _ in self]})"

    def __call__(self, mapping: dict = None, / , **kwargs: Any) -> Schema:
        if mapping is None and not kwargs:
            if not self._database.shape[self._name].registry:
                self._database._reflect_schema(self._name)
        else:
            super().__call__(mapping, **kwargs)

        return self

    def __getattr__(self, attr: str) -> Model:
        if not attr.startswith("_"):
            self._database._reflect_object(ObjectName(stem=attr, schema=self._name))

        try:
            return super().__getattribute__(attr)
        except AttributeError:
            raise AttributeError(f"{type(self).__name__} '{self._name}' of {type(self._database).__name__} '{self._database.name}' has no object '{attr}'.")

    def _refresh(self, shape: SchemaShape) -> BaseSchema:
        return self


class Schema(BaseSchema):
    def _refresh(self, shape: SchemaShape) -> Schema:
        self({name.stem: ObjectProxy(name, parent=self) for name in shape.objects})
        return self


class TableSchema(BaseSchema):
    def _refresh(self, shape: SchemaShape) -> TableSchema:
        self({name.stem: ObjectProxy(name, parent=self) for name in shape.tables})
        return self


class ViewSchema(BaseSchema):
    def _refresh(self, shape: SchemaShape) -> ViewSchema:
        self({name.stem: ObjectProxy(name, parent=self) for name in shape.views})
        return self


class ObjectProxy:
    def __init__(self, name: ObjectName, parent: BaseSchema) -> None:
        self.name, self.parent, self.database = name, parent, parent._database
        self.registry = self.database.shape.schemas[parent._name].registry

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name='{self.name.name}', stem='{self.name.stem}', schema='{self.name.schema}')"

    def __call__(self) -> Model:
        if model := self.registry.get(self.name.stem):
            return model
        else:
            self.database._reflect_object(name=self.name)
            return self.registry.get(self.name.stem)


class BaseSchemas(NameSpace):
    """A NameSpace class representing a set of database schemas. Individual schemas can be accessed with either attribute or item access. If a schema isn't already cached an attempt will be made to reflect it."""
    _schema_constructor = BaseSchema

    def __init__(self, database: Database) -> None:
        self._database = database

    def __repr__(self) -> str:
        return f"""{type(self).__name__}(num_schemas={len(self)}, schemas=[{", ".join([f"{type(schema).__name__}(name='{name}', tables={len(schema)})" for name, schema in self])}])"""

    def __call__(self, mapping: dict = None, / , **kwargs: Any) -> Schema:
        if mapping is None and not kwargs:
            self._database._reflect_database()
        else:
            super().__call__(mapping, **kwargs)

        return self

    def __getattr__(self, attr: str) -> Schema:
        if attr == "__none__":
            return self[self._database.default_schema]

        if not attr.startswith("_"):
            self._database.router.refresh_accessors()

        try:
            return super().__getattribute__(attr)
        except AttributeError:
            raise AttributeError(f"{type(self._database).__name__} '{self._database.name}' has no schema '{attr}'.")

    def _refresh(self, shape: DatabaseShape) -> None:
        self({schema_name.name: self._schema_constructor(name=schema_name, parent=self)._refresh(schema_shape) for schema_name, schema_shape in shape.schemas.items()})


class Schemas(BaseSchemas):
    _schema_constructor = Schema


class TableSchemas(BaseSchemas):
    _schema_constructor = TableSchema


class ViewSchemas(BaseSchemas):
    _schema_constructor = ViewSchema


class SchemaRouter:
    def __init__(self, database: Database) -> None:
        self.database = database
        self.accessors: List[Schemas] = []

    def register(self, accessor: Schemas) -> SchemaRouter:
        self.accessors.append(accessor)
        return self

    def register_all(self, accessors: Collection[Schemas]) -> SchemaRouter:
        self.accessors += accessors
        return self

    def refresh_accessors(self) -> None:
        for accessor in self.accessors:
            accessor._refresh(self.database.shape)

    def remove_object_if_exists(self, name: ObjectName) -> None:
        for accessor in self.accessors:
            try:
                del accessor[name.schema][name.stem]
            except Exception:
                pass



from __future__ import annotations

import os
import warnings

from typing import TYPE_CHECKING

from django.apps import AppConfig
from django import db
from django.conf import settings

from subtypes import Dict_

from sqlhandler.sql import Sql


if TYPE_CHECKING:
    from .sql import DjangoSql


class NullOp:
    pass


class SqlConfig(AppConfig):
    name, Sql, connections = "sqlhandler", Sql, Dict_()
    sql: DjangoSql = None

    settings = Dict_(
        {
            "SCHEMAS": [None],
            "ENGINES": {
                "sqlite3": "sqlite",
                "mysql": "mysql",
                "postgresql": "postgresql",
                "postgresql_psycopg2": "postgresql+psycopg2",
                "oracle": "oracle",
            },
            "MODEL_MIXIN": NullOp,
        }
    )

    def ready(self) -> None:
        self.setup()

    def setup(self) -> None:
        import sqlhandler.django as root

        self.settings.update(getattr(settings, "SQLHANDLER_SETTINGS", {}))
        for connection in db.connections.databases:
            self.connections[connection] = self.initialize_database(connection=connection)

        type(self).sql = root.sql = self.connections.default or None

    def initialize_database(self, connection: str):
        from .sql import DjangoSql

        try:
            return DjangoSql(connection=connection)
        except Exception as ex:
            import traceback
            print(f"The following exception caused sqlhandler.django to fail to start but was suppressed:\n\n{traceback.format_exc()}")
            return None



from __future__ import annotations

from typing import Any, Union, Dict, TYPE_CHECKING, Type, cast

import sqlalchemy as alch
from sqlalchemy import Column, true, null, func
from sqlalchemy import types, event
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.orm import backref
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.orm.mapper import Mapper

from sqlalchemy.ext.declarative import declared_attr, DeclarativeMeta

from subtypes import Str

from .field import SubtypesDateTime
from .misc import absolute_namespace, CreateTableAccessor
from .relationship import Relationship
from .query import Query
from .table import Table

if TYPE_CHECKING:
    from sqlhandler.database import Metadata


class ModelMeta(DeclarativeMeta):
    _registry = set()

    __table__ = None
    __table_cls__ = Table
    metadata: Metadata

    def __new__(mcs, name: str, bases: tuple, namespace: dict) -> ModelMeta:
        if name == "BaseModel" and not bases:
            return type(name, bases, namespace)

        abs_ns = absolute_namespace(bases=bases, namespace=namespace)

        if relationships := {key: val for key, val in abs_ns.items() if isinstance(val, Relationship)}:
            table_name = type(name, (), abs_ns).__tablename__
            for attribute, relationship in relationships.items():
                relationship.build(table_name=table_name, bases=bases, namespace=namespace, attribute=attribute)

        cls = cast(Type[BaseModel], type.__new__(mcs, name, bases, namespace))
        cls._registry.add(cls)
        return cls

    def __repr__(cls) -> str:
        return cls.__name__ if cls.__table__ is None else f"{cls.__name__}({', '.join([f'{col.key}={type(col.type).__name__}' for col in cls.__table__.columns])})"

    def __getitem__(cls, item: str) -> InstrumentedAttribute:
        return getattr(cls, item)

    @property
    def query(cls: ModelMeta) -> Query:
        """Create a new Query operating on this class."""
        return cls.metadata.sql.session.query(cls)

    @property
    def create(cls: ModelMeta) -> CreateTableAccessor:
        """Create the table mapped to this class."""
        return CreateTableAccessor(cls)

    @property
    def c(cls: ModelMeta) -> ImmutableColumnCollection:
        """Access the columns (or a specific column if 'colname' is specified) of the underlying table."""
        return cls.__table__.c

    def alias(cls: ModelMeta, name: str, **kwargs: Any) -> AliasedClass:
        """Create a new class that is an alias of this one, with the given name."""
        return alch.orm.aliased(cls, name=name, **kwargs)

    def drop(cls: ModelMeta) -> None:
        """Drop the table mapped to this class."""
        cls.metadata.sql.database.drop_table(cls)


class BaseModel(metaclass=ModelMeta):
    """Custom base class for declarative and automap bases to inherit from. Represents a mapped table in a sql database."""
    __table__: Table
    metadata: Metadata
    __mapper__: Mapper

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{col.name}={repr(getattr(self, col.name))}' for col in type(self).__table__.columns])})"

    def insert(self) -> BaseModel:
        """Emit an insert statement for this object against this model's underlying table."""
        self.metadata.sql.session.add(self)
        return self

    def update(self, argdeltas: Dict[Union[str, InstrumentedAttribute], Any] = None, /, **update_kwargs: Any) -> BaseModel:
        """
        Emit an update statement against database record represented by this object in this model's underlying table.
        This method positionally accepts a dict where the keys are the model's class attributes (of type InstrumentedAttribute) and the values are the values to update to.
        Alternatively, if the column names are known they may be set using keyword arguments. Raises AttributeError if invalid keys are provided.
        """
        updates: Dict[str, Any] = {}

        clean_argdeltas = {} if argdeltas is None else {(name if isinstance(name, str) else name.key): val for name, val in argdeltas.items()}
        updates.update(clean_argdeltas)
        updates.update(update_kwargs)

        if difference := (set(updates) - set([attr.key for attr in self.__mapper__.all_orm_descriptors])):
            raise AttributeError(f"""Cannot perform update, '{type(self).__name__}' object has no attribute(s): {", ".join([f"'{unknown}'" for unknown in difference])}.""")

        if clean_argdeltas and update_kwargs:
            if intersection := (set(clean_argdeltas) & set(update_kwargs)):
                raise AttributeError(f"""Attribute(s) {", ".join([f"'{dupe}'" for dupe in intersection])} was/were provided twice.""")

        for key, val in updates.items():
            setattr(self, key, val)

        return self

    def delete(self) -> BaseModel:
        """Emit a delete statement for this object against this model's underlying table."""
        self.metadata.sql.session.delete(self)
        return self

    def clone(self, argdeltas: Dict[Union[str, InstrumentedAttribute], Any] = None, **update_kwargs: Any) -> BaseModel:
        """Create a clone (new primary_key, but copies of all other attributes) of this object in the detached state. Model.insert() will be required to persist it to the database."""
        valid_cols = [col.name for col in self.__table__.columns if col.name not in self.__table__.primary_key.columns]
        return type(self)(**{col: getattr(self, col) for col in valid_cols}).update(argdeltas, **update_kwargs)


class Model(BaseModel):
    @declared_attr
    def __table_args__(cls):
        schema = cls.__dict__.get("__schema__", cls.metadata.sql.database.default_schema)
        return dict(schema=schema, is_declarative=True)


class AutoModel(Model):
    @declared_attr
    def __tablename__(cls):
        return str(Str(cls.__name__).case.snake())

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(50), nullable=True, server_default=null())

    @declared_attr
    def created(cls):
        return Column(SubtypesDateTime, nullable=False, server_default=func.NOW())

    @declared_attr
    def modified(cls):
        return Column(SubtypesDateTime, nullable=False, server_default=func.NOW(), onupdate=func.NOW())

    @declared_attr
    def active(cls):
        return Column(types.Boolean, nullable=False, server_default=true())


class ReflectedModel(BaseModel):
    pass


reserved_colnames = set(dir(ModelMeta))


@event.listens_for(Table, "column_reflect")
def on_column_reflect(inspector, table, column_info):
    if (key := column_info["name"]) in reserved_colnames:
        column_info["key"] = f"{key}_"



from __future__ import annotations

import os

from typing import Any, Dict, TYPE_CHECKING


import numpy as np
import pandas as pd

import sqlalchemy as alch
from sqlalchemy.orm import backref, relationship
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.dialects import mssql

from subtypes import Frame, Enum
from pathmagic import File, PathLike
from miscutils import cached_property
from iotools.misc.serializer import LostObject

from sqlhandler.custom import ModelMeta, Model, AutoModel, ReflectedModel, Query, Session, ForeignKey, Relationship, SubtypesDateTime, SubtypesDate, BitLiteral, Select, Update, Insert, Delete, SelectInto
from sqlhandler.database.schema import TableSchemas, ViewSchemas
from sqlhandler.utils import StoredProcedure, Script, SqlLog
from sqlhandler.database import Database, Metadata, Schemas, Schema
from sqlhandler.utils import Config, Url

if TYPE_CHECKING:
    from alembic.operations import Operations


class Sql:
    """
    Provides access to the complete sqlalchemy API, with custom functionality added for logging and pandas integration. Handles authentication through config settings.
    The custom expression classes provided have additional useful methods and are modified by the 'autocommit' attribute to facilitate human-supervised queries.
    The custom query class provided by the Alchemy object's 'session' attribute also has additional methods. Many commonly used sqlalchemy objects are bound to this object as attributes for easy access.
    The 'Sql.orm' and 'Sql.objects' attributes provide access via attribute or item access to the reflected database models and underlying table objects, respectively.
    """

    class IfExists(Enum):
        FAIL, REPLACE, APPEND = "fail", "replace", "append"

    class Constructors:
        ModelMeta, Model, AutoModel, ReflectedModel = ModelMeta, Model, AutoModel, ReflectedModel
        Database, Metadata, Schemas, Schema = Database, Metadata, Schemas, Schema
        Config, Query, Session = Config, Query, Session
        StoredProcedure, Script = StoredProcedure, Script
        Config, Frame = Config, Frame

    class Settings:
        cache_metadata = reflect_tables = reflect_views = True
        eager_reflection = False
        lazy_schemas = {"information_schema"}

    def __init__(self, connection: str = None, database: str = None, log: File = None, autocommit: bool = False) -> None:
        self.constructors, self.settings = self.Constructors(), self.Settings()
        self.config = self.constructors.Config()

        self.engine = self._create_engine(connection=connection, database=database)
        self.engine.connect()

        self.session = self.constructors.Session(bind=self.engine, query_cls=self.constructors.Query)
        self.database = self.constructors.Database(self)

        self.log, self.autocommit = log, autocommit

        self.Select, self.SelectInto, self.Update, self.Insert, self.Delete = Select, SelectInto, Update, Insert, Delete
        self.StoredProcedure, self.Script = self.constructors.StoredProcedure.from_sql(self), self.constructors.Script.from_sql(self)

        self.text, self.literal = alch.text, alch.literal
        self.AND, self.OR, self.CAST, self.CASE, self.TRUE, self.FALSE = alch.and_, alch.or_, alch.cast, alch.case, alch.true(), alch.false()

        self.Table, self.Column, self.ForeignKey, self.Index, self.CheckConstraint, self.Relationship, self.relationship, self.backref = alch.Table, alch.Column, ForeignKey, alch.Index, alch.CheckConstraint, Relationship, relationship, backref
        self.Datetime, self.Date = SubtypesDateTime, SubtypesDate
        self.type, self.func, self.sqlalchemy = alch.types, alch.func, alch

    def __repr__(self) -> str:
        return f"{type(self).__name__}(engine={repr(self.engine)}, database={repr(self.database)})"

    def __len__(self) -> int:
        return len(self.database.meta.tables)

    def __enter__(self) -> Sql:
        self.session.rollback()
        return self

    def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
        self.session.commit() if ex_type is None else self.session.rollback()

    def __getstate__(self) -> dict:
        return {"engine": LostObject(self.engine), "database": LostObject(self.database), "autocommit": self.autocommit}

    def __setstate__(self, attrs: dict) -> None:
        self.__dict__ = attrs

    @property
    def Model(self) -> Model:
        """Custom base class for declarative and automap bases to inherit from. Represents a mapped table in a sql database."""
        return self.database.model

    @property
    def AutoModel(self) -> AutoModel:
        return self.database.auto_model

    @property
    def objects(self) -> Schemas:
        """Property controlling access to all database objects."""
        return self.database.objects

    @property
    def tables(self) -> TableSchemas:
        """Property controlling access to mapped tables."""
        return self.database.tables

    @property
    def views(self) -> ViewSchemas:
        """Property controlling access to mapped views."""
        return self.database.views

    @cached_property
    def operations(self) -> Operations:
        """Property controlling access to alembic operations."""
        from alembic.runtime.migration import MigrationContext
        from alembic.operations import Operations

        return Operations(MigrationContext.configure(self.engine.connect()))

    @property
    def log(self) -> SqlLog:
        """Property controlling access to a special SqlLog class. When setting, a simple file path should be used and an appropriate SqlLog will be created."""
        return self._log

    @log.setter
    def log(self, val: File) -> None:
        self._log = SqlLog(path=val)

    def initialize_log(self, name: str, location: PathLike = None) -> SqlLog:
        """Instantiates a matt.log.SqlLog object from a name and a dirpath, and binds it to this object's 'log' attribute. If 'active' argument is 'False', this method does nothing."""
        self._log = SqlLog.from_details(log_name=name, log_dir=location, active=False)
        return self._log

    # Conversion Methods

    def query_to_frame(self, query: Query, labels: bool = False) -> Frame:
        """Convert sqlalchemy.orm.Query object to a pandas DataFrame. Optionally apply table labels to columns and/or print an ascii representation of the DataFrame."""
        query = query.with_labels() if labels else query

        result = self.session.execute(query.statement)
        cols = [col[0] for col in result.cursor.description]
        frame = self.constructors.Frame(result.fetchall(), columns=cols)

        return frame

    def plaintext_query_to_frame(self, query: str) -> Frame:
        """Convert plaintext SQL to a pandas DataFrame. The SQL statement must be a SELECT that returns rows."""
        return self.constructors.Frame(pd.read_sql_query(query, self.engine))

    def table_to_frame(self, table: str, schema: str = None) -> Frame:
        """Reads the target table or view (from the specified schema) into a pandas DataFrame."""
        return self.constructors.Frame(pd.read_sql_table(table, self.engine, schema=schema))

    def excel_to_table(self, filepath: os.PathLike, table: str = "temp", schema: str = None, if_exists: Sql.IfExists = IfExists.FAIL, primary_key: str = "id", **kwargs: Any) -> Model:
        """Bulk insert the content of the target '.xlsx' file to the specified table."""
        return self.frame_to_table(dataframe=self.constructors.Frame.from_excel(filepath, **kwargs), table=table, schema=schema, if_exists=if_exists, primary_key=primary_key)

    def frame_to_table(self, dataframe: pd.DataFrame, table: str, schema: str = None, if_exists: Sql.IfExists = IfExists.FAIL, primary_key: str = "id") -> Model:
        """Bulk insert the content of a pandas DataFrame to the specified table."""
        dataframe = self.constructors.Frame(dataframe)

        has_identity_pk = False
        if primary_key is None:
            dataframe.reset_index(inplace=True)
            primary_key = dataframe.iloc[:, 0].name
        else:
            if primary_key in dataframe.columns:
                dataframe.set_index(primary_key, inplace=True)
            else:
                has_identity_pk = True
                dataframe.reset_index(inplace=True, drop=True)
                dataframe.index.names = [primary_key]
                dataframe.index += 1

            dataframe.reset_index(inplace=True)

        dtypes = self._sql_dtype_dict_from_frame(dataframe)
        if has_identity_pk:
            dtypes[primary_key] = alch.types.INT

        dataframe.to_sql(engine=self.engine, name=table, if_exists=if_exists, index=False, primary_key=primary_key, schema=schema, dtype=dtypes)

        self.database._sync_with_db()
        return self.tables[schema][table]()

    def orm_to_frame(self, orm_objects: Any) -> Frame:
        """Convert a homogeneous list of sqlalchemy.orm instance objects (or a single one) to a pandas DataFrame."""
        if not isinstance(orm_objects, list):
            orm_objects = [orm_objects]

        if not all([type(orm_objects[0]) == type(item) for item in orm_objects]):
            raise TypeError("All sqlalchemy.orm mapped objects passed into this function must have the same type.")

        cols = [col.name for col in list(type(orm_objects[0]).__table__.columns)]
        vals = [[getattr(item, col) for col in cols] for item in orm_objects]

        return self.constructors.Frame(vals, columns=cols)

    # Private internal methods

    def _create_engine(self, connection: str, database: str) -> alch.engine.base.Engine:
        url = self._create_url(connection=connection, database=database)

        engine = alch.create_engine(str(url), echo=False, dialect=self._customize_dialect(url.get_dialect()()))
        engine.sql = self

        return engine

    def _create_url(self, connection: str, database: str) -> Url:
        return self.config.generate_url(connection=connection, database=database)

    def _customize_dialect(self, dialect: DefaultDialect) -> DefaultDialect:
        dialect.colspecs.update(
            {
                alch.types.DateTime: SubtypesDateTime,
                alch.types.DATETIME: SubtypesDateTime,
                alch.types.Date: SubtypesDate,
                alch.types.DATE: SubtypesDate,
            }
        )

        if isinstance(dialect, mssql.dialect):
            dialect.supports_multivalues_insert = True
            dialect.colspecs.update({alch.dialects.mssql.BIT: BitLiteral})

        return dialect

    @staticmethod
    def _sql_dtype_dict_from_frame(frame: Frame) -> Dict[str, Any]:
        def isnull(val: Any) -> bool:
            return val is None or np.isnan(val)

        def sqlalchemy_dtype_from_series(series: pd.code.series.Series) -> Any:
            if series.dtype.name in ["int64", "Int64"]:
                if not (nums := [num for num in series if not isnull(num)]):
                    return alch.types.Integer
                else:
                    minimum, maximum = min(nums), max(nums)

                    if 0 <= minimum and maximum <= 255:
                        return alch.dialects.mssql.TINYINT
                    elif -2**15 <= minimum and maximum <= 2**15:
                        return alch.types.SmallInteger
                    elif -2**31 <= minimum and maximum <= 2**31:
                        return alch.types.Integer
                    else:
                        return alch.types.BigInteger
            elif series.dtype.name == "object":
                return alch.types.String(int((series.fillna("").astype(str).str.len().max()//50 + 1)*50))
            else:
                raise TypeError(f"Don't know how to process column type '{series.dtype}' of '{series.name}'.")

        return {name: sqlalchemy_dtype_from_series(col) for name, col in frame.infer_objects().iteritems() if col.dtype.name in ["int64", "Int64", "object"]}



from __future__ import annotations

from typing import Any, Callable, TYPE_CHECKING, Iterator, Tuple

from django.apps import apps

from maybe import Maybe
from subtypes import Str

from .config import SqlConfig

if TYPE_CHECKING:
    from .sql import DjangoSql


class DjangoApp(SqlConfig.Sql.Constructors.Schema):
    pass


class DjangoApps(SqlConfig.Sql.Constructors.Schemas):
    schema_constructor = DjangoApp

    def __repr__(self) -> str:
        return f"""{type(self).__name__}(num_apps={len(self)}, apps=[{", ".join([f"{type(schema).__name__}(name='{schema._name}', tables={len(schema) if schema._ready else '?'})" for name, schema in self])}])"""

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        return super().__iter__()


class DjangoDatabase(SqlConfig.Sql.Constructors.Database):
    def __init__(self, sql: DjangoSql) -> None:
        self.django_mappings = {model._meta.db_table: model for models in apps.all_models.values() for model in models.values()}
        self.sqlhandler_mappings = {}
        super().__init__(sql=sql)

        self.django = DjangoApps(database=self)
        self._hierarchize()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.name)}, django={repr(self.django)})"

    def _hierarchize(self) -> None:
        for app, models in apps.all_models.items():
            self.django[app] = schema = self.django.schema_constructor(name=app, parent=self.django)
            schema._ready = True

            for name, model in models.items():
                if (model := self.shape[self.default_schema].registry.get((table_name := model._meta.db_table))) is not None:
                    self.sqlhandler_mappings[table_name] = schema[name] = model

    def _scalar_name(self) -> Callable:
        def scalar_name(base: Any, local_cls: Any, referred_cls: Any, constraint: Any) -> str:
            return Maybe(self.django_mappings)[referred_cls.__name__]._meta.model_name.else_(referred_cls.__name__)

        return scalar_name

    def _collection_name(self) -> Callable:
        def collection_name(base: Any, local_cls: Any, referred_cls: Any, constraint: Any) -> str:
            real_name = Maybe(self.django_mappings)[referred_cls.__name__]._meta.model_name.else_(referred_cls.__name__)
            return Str(real_name).case.plural()

        return collection_name
